﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Reflection;
using PropertyGridEditors.Converters;

namespace PropertyGridEditors.Editors
{
    /// <summary>
    /// Describes a numeric editor.
    /// </summary>
    /// <!-- DPE -->
    public partial class NumericUpDown : UserControl
    {
        #region Fields

        /// <summary>
        /// List of the types owned by this editor.
        /// </summary>
        private List<Type> mOwnedTypes;

        /// <summary>
        /// Old the previous value when the textbox get focused.
        /// Manly usefull when copy and paste is performed.
        /// </summary>
        private Object mPreviousValue;

        /// <summary>
        /// Converter used to convert the edited value to a string.
        /// </summary>
        private ToStringConverter mTextBoxConverter;

        #endregion // Fields.

        #region Dependency properties

        /// <summary>
        /// The dependency property which owned the ValueType property.
        /// </summary>
        public static readonly DependencyProperty ValueTypeProperty = DependencyProperty.Register("ValueType",
                typeof(Type), typeof(NumericUpDown),
                new PropertyMetadata(typeof(System.Int32), null, new CoerceValueCallback(CoerceTypeValue)));

        /// <summary>
        /// The dependency property which owned the Precision property.
        /// </summary>
        public static readonly DependencyProperty PrecisionProperty = DependencyProperty.Register("Precision",
                typeof(Int32), typeof(NumericUpDown),
                new PropertyMetadata(0, new PropertyChangedCallback(OnPrecisionChanged)));

        /// <summary>
        /// The dependency property which owned the Maximum property.
        /// </summary>
        public static readonly DependencyProperty MaxValueProperty = DependencyProperty.Register("MaxValue",
                typeof(Object), typeof(NumericUpDown),
                new PropertyMetadata(null, new PropertyChangedCallback(OnDependencyValueChanged),
                    new CoerceValueCallback(CoerceMaxValue)));

        /// <summary>
        /// The dependency property which owned the Minimum property.
        /// </summary>
        public static readonly DependencyProperty MinValueProperty = DependencyProperty.Register("MinValue",
                typeof(Object), typeof(NumericUpDown),
                new PropertyMetadata(null, new PropertyChangedCallback(OnDependencyValueChanged),
                    new CoerceValueCallback(CoerceMinValue)));

        /// <summary>
        /// The dependency property which owned the StepValue property.
        /// </summary>
        public static readonly DependencyProperty StepValueProperty = DependencyProperty.Register("StepValue",
                typeof(Object), typeof(NumericUpDown),
                new PropertyMetadata(null, new PropertyChangedCallback(OnDependencyValueChanged),
                    new CoerceValueCallback(CoerceValueOnType)));

        /// <summary>
        /// The dependency property which owned the Value property.
        /// </summary>
        public static readonly DependencyProperty ValueProperty = DependencyProperty.Register("Value",
                typeof(Object), typeof(NumericUpDown),
                new PropertyMetadata(null, new PropertyChangedCallback(OnDependencyValueChanged),
                    new CoerceValueCallback(CoerceValueOnType)));

        /// <summary>
        /// The dependency property which owned the Value property.
        /// </summary>
        public static readonly DependencyProperty LoopProperty = DependencyProperty.Register("Loop",
                typeof(Boolean), typeof(NumericUpDown), new PropertyMetadata(false, null));

        #endregion // Dependency properties.

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the ButtonUpDown class.
        /// </summary>
        public NumericUpDown()
        {
            try
            {
                InitializeComponent();
            }
            catch (Exception lEx)
            {
                // Throwing the exception.
                throw new Exception("NumericUpDown creation error !!!", lEx);
            }

            // Defining the owned types.
            mOwnedTypes = new List<Type>();
            mOwnedTypes.Add(typeof(System.Int32));
            mOwnedTypes.Add(typeof(System.Single));
            mOwnedTypes.Add(typeof(System.Double));

            // Handling textbox events.
            mTextBox.PreviewTextInput += new TextCompositionEventHandler(OnTextBoxPreviewTextInput);
            mTextBox.GotFocus += new RoutedEventHandler(OnTextBoxGotFocus);
            mTextBox.LostFocus += new RoutedEventHandler(OnTextBoxLostFocus);

            // Handing incrementation and decrementation.
            mUpButton.Click += new RoutedEventHandler(OnUpButtonClick);
            mDownButton.Click += new RoutedEventHandler(OnDownButtonClick);

            // Handling the conversion to and from the textbox content string.
            // Binding is done dynamically and not in the XAML to be able to handle the textbox 
            // display format depending on the type in the converter.
            Binding lEditionBinding = new Binding("Value");
            lEditionBinding.Source = this;
            lEditionBinding.Mode = BindingMode.TwoWay;
            mTextBoxConverter = new ToStringConverter();
            lEditionBinding.Converter = mTextBoxConverter;
            lEditionBinding.UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged;
            mTextBox.SetBinding(TextBox.TextProperty, lEditionBinding);
        }

        #endregion // Constructors.

        #region Properties

        /// <summary>
        /// Gets or sets the editor value.
        /// </summary>
        public Object Value
        {
            get
            {
                return GetValue(ValueProperty);
            }
            set
            {
                SetValue(ValueProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the editor maximum value.
        /// </summary>
        public Object MaxValue
        {
            get
            {
                return GetValue(MaxValueProperty);
            }
            set
            {
                SetValue(MaxValueProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the editor minimum value.
        /// </summary>
        public Object MinValue
        {
            get
            {
                return GetValue(MinValueProperty);
            }
            set
            {
                SetValue(MinValueProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the step used to increment or decrement the current value.
        /// </summary>
        public Object StepValue
        {
            get
            {
                return GetValue(StepValueProperty);
            }
            set
            {
                SetValue(StepValueProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the loop property.
        /// </summary>
        public Boolean Loop
        {
            get
            {
                return (Boolean)GetValue(LoopProperty);
            }
            set
            {
                SetValue(LoopProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the type of the value the editor is editing.
        /// </summary>
        public Type ValueType
        {
            get
            {
                return (Type)GetValue(ValueTypeProperty);
            }
            set
            {
                SetValue(ValueTypeProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the precision of the value.
        /// </summary>
        public Int32 Precision
        {
            get
            {
                return (Int32)GetValue(PrecisionProperty);
            }
            set
            {
                SetValue(PrecisionProperty, value);

                // Updating the format in the converter.
                String lFormat = "{0:0";
                if
                    (value > 0)
                {
                    lFormat += ".";
                }
                for
                    (Int32 lIter = 0; lIter < value; lIter++)
                {
                    lFormat += "0";
                }
                lFormat += "}";

                mTextBoxConverter.StringFormat = lFormat;
            }
        }

        #endregion // Properties.

        #region Methods

        /// <summary>
        /// Verifies if the type can be hendled by the editor.
        /// </summary>
        /// <param name="pType">Teh type to validate.</param>
        /// <returns>Return true if the type is handled, false otherwise.</returns>
        public Boolean IsTypeHandled(Type pType)
        {
            return mOwnedTypes.Contains(pType);
        }

        /// <summary>
        /// Method called after a dependency property concerning a value is modified.
        /// </summary>
        /// <param name="pDependencyObject">The dependency object.</param>
        /// <param name="pArgs">The event arguments.</param>
        private static void OnPrecisionChanged(DependencyObject pDependencyObject, DependencyPropertyChangedEventArgs pArgs)
        {
            NumericUpDown lControl = pDependencyObject as NumericUpDown;
            if
                (lControl != null)
            {
                // Updating the precision for the converter.
                try
                {
                    Int32 lPrecision = Convert.ToInt32(pArgs.NewValue);
                    lControl.Precision = lPrecision;
                }
                catch
                { }
            }
        }

        /// <summary>
        /// Method called after a dependency property concerning a value is modified.
        /// </summary>
        /// <param name="pDependencyObject">The dependency object.</param>
        /// <param name="pArgs">The event arguments.</param>
        private static void OnDependencyValueChanged(DependencyObject pDependencyObject, DependencyPropertyChangedEventArgs pArgs)
        {
            NumericUpDown lControl = pDependencyObject as NumericUpDown;
            if
                (lControl != null)
            {
                // Replace the value in the range if it is out of bound.
                lControl.ConstrainValueToRange();
            }
        }

        /// <summary>
        /// Validate the current value depending on the type.
        /// </summary>
        /// <param name="pDependencyObject">The dependency object.</param>
        /// <param name="pValue">The value to validate.</param>
        /// <returns>The value if it has the good type, null otherwise.</returns>
        private static Object CoerceValueOnType(DependencyObject pDependencyObject, Object pValue)
        {
            NumericUpDown lControl = pDependencyObject as NumericUpDown;
            if
                (lControl != null)
            {
                if
                    (pValue != null
                    && pValue.GetType() != lControl.ValueType)
                {
                    // Disabling the edition.
                    lControl.IsEnabled = false;

                    return null;
                }
            }

            return pValue;
        }

        /// <summary>
        /// Validate the max value has the good type and is bigger than the min value.
        /// </summary>
        /// <param name="pDependencyObject">The dependency object.</param>
        /// <param name="pValue">The value to validate.</param>
        /// <returns>The value if it has the good type and is bigger than the min value, null otherwise.</returns>
        private static Object CoerceMaxValue(DependencyObject pDependencyObject, Object pValue)
        {
            // Vaditating the value type.
            IComparable lMaxValue = NumericUpDown.CoerceValueOnType(pDependencyObject, pValue) as IComparable;
            
            // Validating the min and max order.
            if
                (lMaxValue != null)
            {
                NumericUpDown lControl = pDependencyObject as NumericUpDown;
                if
                    (lControl != null)
                {
                    IComparable lMinValue = lControl.MinValue as IComparable;
                    if
                        (lMinValue != null
                        && lMaxValue.CompareTo(lMinValue) < 0)
                    {
                        // The MaxValue is lower than the MinValue.
                        return lMinValue;
                    }
                }
            }

            return lMaxValue;
        }

        /// <summary>
        /// Validate the min value has the good type and is lower than the max value.
        /// </summary>
        /// <param name="pDependencyObject">The dependency object.</param>
        /// <param name="pValue">The value to validate.</param>
        /// <returns>The value if it has the good type and is lower than the max value, null otherwise.</returns>
        private static Object CoerceMinValue(DependencyObject pDependencyObject, Object pValue)
        {
            // Vaditating the value type.
            IComparable lMinValue = NumericUpDown.CoerceValueOnType(pDependencyObject, pValue) as IComparable;

            // Validating the min and max order.
            if
                (lMinValue != null)
            {
                NumericUpDown lControl = pDependencyObject as NumericUpDown;
                if
                    (lControl != null)
                {
                    IComparable lMaxValue = lControl.MaxValue as IComparable;
                    if
                        (lMaxValue != null
                        && lMinValue.CompareTo(lMaxValue) > 0)
                    {
                        // The MaxValue is lower than the MinValue.
                        return lMaxValue;
                    }
                }
            }

            return lMinValue;
        }

        /// <summary>
        /// Verify if the given type is handled by the editor.
        /// </summary>
        /// <param name="pDependencyObject">The dependency object.</param>
        /// <param name="pType">The type to validate.</param>
        /// <returns>The type if it is handled, or Int32 otherwise.</returns>
        private static Object CoerceTypeValue(DependencyObject pDependencyObject, Object pValue)
        {
            Type lReturnType = pValue as Type;
            NumericUpDown lControl = pDependencyObject as NumericUpDown;
            if
                (lControl != null)
            {
                if
                    (lControl.IsTypeHandled(lReturnType))
                {
                    return lReturnType;
                }
            }

            return typeof(System.Int32);
        }

        /// <summary>
        /// Method called when an input is about to be entered.
        /// </summary>
        /// <param name="pSender">The object sender.</param>
        /// <param name="pArgs">The event arguments.</param>
        private void OnTextBoxPreviewTextInput(Object pSender, TextCompositionEventArgs pArgs)
        {
            if
                (!IsNumericInput(pArgs.Text))
            {
                pArgs.Handled = true;
                return;
            }
        }

        /// <summary>
        /// Method called when the textbox got the focus.
        /// </summary>
        /// <param name="pSender">The object sender.</param>
        /// <param name="pArgs">The event arguments.</param>
        private void OnTextBoxGotFocus(Object pSender, RoutedEventArgs pArgs)
        {
            // Saving the previous value.
            mPreviousValue = Value;
        }

        /// <summary>
        /// Method called when the textbox lost the focus.
        /// </summary>
        /// <param name="pSender">The object sender.</param>
        /// <param name="pArgs">The event arguments.</param>
        private void OnTextBoxLostFocus(Object pSender, RoutedEventArgs pArgs)
        {
            try
            {
                // Verify the value is of type ValueType.
                Object lNumericValue = System.Convert.ChangeType(mTextBox.Text, ValueType);

                // Replace the value in the range if it is out of bound.
                ConstrainValueToRange();
            }
            catch
                (Exception /*pException*/)
            {
                // The value is not a numeric type, reinitializing the textbox 
                // with the old value.
                mTextBox.Text = mPreviousValue.ToString();
            }
        }

        /// <summary>
        /// Method called when the down button is clicked.
        /// </summary>
        /// <param name="pSender">The object sender.</param>
        /// <param name="pArgs">The event arguments.</param>
        private void OnDownButtonClick(Object pSender, RoutedEventArgs pArgs)
        {
            if
                (Value != null
                && StepValue != null
                )
            {
                // TODO : Find a way to make it generic...
                if
                    (ValueType == typeof(System.Int32))
                {
                    Value = Convert.ToInt32(Value) - Convert.ToInt32(StepValue);
                }
                else if
                    (ValueType == typeof(System.Single))
                {
                    Value = Convert.ToSingle(Value) - Convert.ToSingle(StepValue);
                }
                else if
                    (ValueType == typeof(System.Double))
                {
                    Value = Convert.ToDouble(Value) - Convert.ToDouble(StepValue);
                }
            }
        }

        /// <summary>
        /// Method called when the up button is clicked.
        /// </summary>
        /// <param name="pSender">The object sender.</param>
        /// <param name="pArgs">The event arguments.</param>
        private void OnUpButtonClick(Object pSender, RoutedEventArgs pArgs)
        {
            if
                (Value != null
                && StepValue != null
                )
            {
                // TODO : Find a way to make it generic...
                if
                    (ValueType == typeof(System.Int32))
                {
                    Value = Convert.ToInt32(Value) + Convert.ToInt32(StepValue);
                }
                else if
                    (ValueType == typeof(System.Single))
                {
                    Value = Convert.ToSingle(Value) + Convert.ToSingle(StepValue);
                }
                else if
                    (ValueType == typeof(System.Double))
                {
                    Value = Convert.ToDouble(Value) + Convert.ToDouble(StepValue);
                }
            }
        }

        /// <summary>
        /// Verify if the caracter if a numeric value.
        /// </summary>
        /// <param name="pText">The caracter to verify.</param>
        /// <returns>True if the caracter is a numeric one.</returns>
        private Boolean IsNumericInput(String pChar)
        {
            if
                (pChar == System.Globalization.NumberFormatInfo.CurrentInfo.CurrencyDecimalSeparator |
                 pChar == System.Globalization.NumberFormatInfo.CurrentInfo.NegativeSign |
                 pChar == System.Globalization.NumberFormatInfo.CurrentInfo.NumberDecimalSeparator |
                 pChar == System.Globalization.NumberFormatInfo.CurrentInfo.PercentDecimalSeparator |
                 pChar == System.Globalization.NumberFormatInfo.CurrentInfo.PositiveSign
                )
            {
                return true;
            }

            foreach
                (char lChar in pChar)
            {
                if
                    (!char.IsDigit(lChar))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Replace the value in the range if it is out of bound.
        /// </summary>
        private void ConstrainValueToRange()
        {
            // Depending on the type, verifying the value is in the range.
            IComparable lValue = Value as IComparable;
            IComparable lMaxValue = MaxValue as IComparable;
            IComparable lMinValue = MinValue as IComparable;

            if
                (lValue != null
                && MaxValue != null
                && MinValue != null
                )
            {
                if
                    (lValue.CompareTo(lMaxValue) > 0)
                {
                    if
                        (Loop)
                    {
                        Value = MinValue;
                    }
                    else
                    {
                        Value = MaxValue;
                    }
                }
                else if
                    (lValue.CompareTo(lMinValue) < 0)
                {
                    if
                        (Loop)
                    {
                        Value = MaxValue;
                    }
                    else
                    {
                        Value = MinValue;
                    }
                }
            }
        }

        #endregion // Methods.
    }
}