﻿using System;
using System.Collections;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;

namespace PropertyGridLib
{
    /// <summary>
    /// This class describes the view model for the property.
    /// </summary>
    /// <!-- NBY -->
    public class PropertyViewModel : PropertyViewModelBase
    {
        #region Fields

        /// <summary>
        /// This field stores the activeness of the view model.
        /// </summary>
        private Boolean mIsEnabled = true;

        /// <summary>
        /// This fiels stores the visibility of the view model.
        /// </summary>
        private Visibility mVisibility = Visibility.Visible;

        #endregion // Fields.

        #region Properties

        /// <summary>
        /// Gets or sets the format string.
        /// </summary>
        public String FormatString
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the height of the editor for the property.
        /// </summary>
        public Double Height
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets a value indicating whether the property should be edited as multiline.
        /// </summary>
        public Boolean AcceptsReturn
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the text wrapping for multiline strings.
        /// </summary>
        public TextWrapping TextWrapping
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets a value indicating whether this instance is enumerable.
        /// </summary>
        public Boolean IsEnumerable
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the instance.
        /// </summary>
        public Object Instance
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets or sets the property descriptor.
        /// </summary>
        public PropertyDescriptor Descriptor
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets or sets a value indicating whether this property is enabled.
        /// </summary>
        public bool IsEnabled
        {
            get
            {
                return mIsEnabled;
            }
            set
            {
                mIsEnabled = value;
             }
        }

        /// <summary>
        /// Gets the visibility of the property.
        /// </summary>
        /// <value>The visibility.</value>
        public Visibility Visibility
        {
            get 
            {
                return mVisibility; 
            }
            set 
            { 
                mVisibility = value; 
            }
        }

        /// <summary>
        /// Gets the first instance if the Instance is an Enumerable, otherwise return the Instance.
        /// </summary>
        /// <value>The first instance.</value>
        public object FirstInstance
        {
            get
            {
                if 
                    (IsEnumerable)
                {
                    IEnumerable lList = Instance as IEnumerable;
                    if
                        (lList == null)
                    {
                        throw new InvalidOperationException("Instance should be a list.");
                    }
                    return lList.Cast<object>().FirstOrDefault();
                }
                return Instance;
            }
        }

        /// <summary>
        /// Gets the instances or a single Instance as an Enumerable.
        /// </summary>
        public IEnumerable Instances
        {
            get
            {
                if (IsEnumerable)
                {
                    IEnumerable lList = Instance as IEnumerable;
                    if
                        (lList == null)
                    {
                        throw new InvalidOperationException("Instance should be a list.");
                    }
                    foreach
                        (Object lObj in lList)
                    {
                        yield return lObj;
                    }
                }
                else
                {
                    yield return Instance;
                }
            }
        }

        /// <summary>
        /// Gets or sets the value of the property.
        /// </summary>
        public Object Value
        {
            get
            {
                Object lValue;
                if 
                    (IsEnumerable)
                {
                    IEnumerable lList = Instance as IEnumerable;
                    if
                        (lList == null)
                    {
                        throw new InvalidOperationException("Instance should be a list.");
                    }
                    lValue = GetValueFromEnumerable(lList);
                    
                }
                else
                {
                    lValue = GetValue(Instance);
                }

                return lValue;
            }
            set
            {
                if 
                    (IsEnumerable)
                {
                    IEnumerable lList = Instance as IEnumerable;
                    if
                        (lList == null)
                    {
                        throw new InvalidOperationException("Instance should be a list.");
                    }
                    foreach 
                        (Object lObject in lList)
                    {
                        SetValue(lObject, value);
                    }
                }
                else
                {
                    SetValue(Instance, value);
                }
            }
        }

        /// <summary>
        /// Gets the name of the property.
        /// </summary>
        public String Name
        {
            get 
            { 
                return Descriptor.Name; 
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is writeable.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance is writeable; otherwise, <c>false</c>.
        /// </value>
        public Boolean IsWriteable
        {
            get 
            { 
                return !IsReadOnly; 
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is read only.
        /// </summary>
        public Boolean IsReadOnly
        {
            get 
            { 
                return Descriptor.IsReadOnly; 
            }
        }

        /// <summary>
        /// Gets the type of the property.
        /// </summary>
        /// <value>The type of the property.</value>
        public Type PropertyType
        {
            get 
            { 
                return Descriptor.PropertyType; 
            }
        }

        /// <summary>
        /// Gets the name of the property.
        /// </summary>
        public String PropertyName
        {
            get 
            { 
                return Descriptor.Name; 
            }
        }

        /// <summary>
        /// Gets the display name.
        /// </summary>
        public override String DisplayString
        {
            get
            {
                return Descriptor.DisplayName;
            }
        }

        /// <summary>
        /// Gets the category.
        /// </summary>
        public String Category
        {
            get 
            { 
                return Descriptor.Category; 
            }
        }

        /// <summary>
        /// Gets the description.
        /// </summary>
        public String Description
        {
            get 
            { 
                return Descriptor.Description; 
            }
        }

        /// <summary>
        /// Gets the property template selector.
        /// </summary>
        public DataTemplateSelector PropertyTemplateSelector
        {
            get
            {
                return PropertyViewModelFactory.Instance;
            }
        }

        #endregion // Properties.

        #region Methods

        /// <summary>
        /// Default constructor.
        /// </summary>
        /// <param name="pInstance">The instance to introspect.</param>
        /// <param name="pPropertyDescriptor">The property descriptor.</param>
        public PropertyViewModel(Object pInstance, PropertyDescriptor pPropertyDescriptor)
        : base()
        {
            Instance = pInstance;
            Descriptor = pPropertyDescriptor;
            if
                (   pPropertyDescriptor != null
                &&  pPropertyDescriptor.Description != String.Empty
                )
            {
                ToolTip = pPropertyDescriptor.Description;
            }
            Height = double.NaN;
        }

        /// <summary>
        /// Subscribes to the ValueChanged event.
        /// </summary>
        public void SubscribeValueChanged()
        {
            SubscribeValueChanged(Descriptor, InstancePropertyChanged);
        }

        /// <summary>
        /// This method is used to subscribe to the "ValueChanged" event.
        /// </summary>
        public void SubscribeValueChanged(PropertyDescriptor pDescriptor, EventHandler pEventHandler)
        {
            pDescriptor.AddValueChanged(Instance, pEventHandler);
        }

        /// <summary>
        /// This method is used to unsubscribe the valu>e changed event.
        /// </summary>
        public void UnsubscribeValueChanged(PropertyDescriptor pDescriptor, EventHandler pEventHandler)
        {
            pDescriptor.RemoveValueChanged(Instance, pEventHandler);
        }

        /// <summary>
        /// This methods notifies when a property of the instance is changed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void InstancePropertyChanged(object sender, EventArgs e)
        {
            // Sending notification when the instance has been changed
            NotifyPropertyChanged("Value");
        }

        /// <summary>
        /// The the current value from an IEnumerable instance
        /// </summary>
        /// <param name="componentList"></param>
        /// <returns></returns>
        protected Object GetValueFromEnumerable(IEnumerable componentList)
        {
            object value = null;
            foreach (object component in componentList)
            {
                object v = GetValue(component);
                if (value == null)
                {
                    value = v;
                }
                if (value != null && v == null)
                {
                    return null;
                }
                if (v != null && !v.Equals(value))
                {
                    return null;
                }
            }
            return value;
        }

        /// <summary>
        /// This method is called when the value is modified.
        /// </summary>
        /// <param name="instance"></param>
        /// <param name="value"></param>
        protected virtual void SetValue(Object pInstance, Object pValue)
        {
            Descriptor.SetValue(pInstance, pValue);
        }

        /// <summary>
        /// This method is called to retrieve the value.
        /// </summary>
        /// <param name="instance"></param>
        /// <returns></returns>
        protected virtual object GetValue(Object pInstance)
        {
            return Descriptor.GetValue(pInstance);
        }

        #endregion // Methods.
    }
}