﻿using System;
using System.ComponentModel;

namespace PropertyGridLib
{
    /// <summary>
    /// This class defines the base class for all other view models (Category, Tab, Properties).
    /// </summary>
    /// <!-- NBY -->
    public abstract class PropertyViewModelBase : INotifyPropertyChanged, IComparable
    {
        #region Events

        /// <summary>
        /// This event is emitted each time a property value is changed.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion // Events.

        #region Properties
        
        /// <summary>
        /// Gets or sets the display string of the view model.
        /// </summary>
        public virtual String DisplayString
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the tool type of the view model.
        /// </summary>
        public Object ToolTip 
        { 
            get; 
            set; 
        }

        /// <summary>
        /// Gets or sets the sort order of the view model.
        /// </summary>
        public Int32 SortOrder 
        { 
            get; 
            set; 
        }

        #endregion // Properties

        #region Methods

        /// <summary>
        /// Default constructor.
        /// </summary>
        protected PropertyViewModelBase()
        {
            SortOrder = int.MinValue;
        }

        /// <summary>
        /// This function overrides ToString behavior.
        /// </summary>
        /// <returns>The "DisplayString" property.</returns>
        public override String ToString()
        {
            return DisplayString;
        }

        /// <summary>
        /// This function must be called each time a property value is changed.
        /// </summary>
        /// <param name="pProperty">The property name.</param>
        protected void NotifyPropertyChanged(String pProperty)
        {
            if 
                (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(pProperty));
            }
        }

        /// <summary>
        /// This function is used to compare the view model.
        /// </summary>
        /// <param name="pObjectToCompare">The object to compare.</param>
        /// <returns>Negative if this is less than pObject</returns>
        public Int32 CompareTo(Object pObjectToCompare)
        {
            return SortOrder.CompareTo(((PropertyViewModelBase)pObjectToCompare).SortOrder);
        }
        
        #endregion // Methods.
    }
}
