﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using System.Windows.Controls;
using System.Windows;

namespace PropertyGridLib
{
    /// <summary>
    /// This class represents the property view model factory.
    /// </summary>
    /// <!-- NBY -->
    public sealed class PropertyViewModelFactory : DataTemplateSelector
    {
        #region Enumeration

        /// <summary>
        /// Attribute type.
        /// </summary>
        public enum AttributeType
        {
            Boolean,
            Int32,
            String
        };

        #endregion

        #region Fields

        /// <summary>
        /// Field which stores the unique instance of the singleton.
        /// </summary>
        static readonly PropertyViewModelFactory msInstance = new PropertyViewModelFactory();

        /// <summary>
        /// This field stores the association between a data type and its view model.
        /// </summary>
        Dictionary<Type, Type> mViewModelByDataType = null;

        /// <summary>
        /// This field stores the association between a view model type and a data template.
        /// </summary>
        Dictionary<Type, String> mDataTemplateByViewModel = null;

        #endregion // Fields.

        #region Methods

        /// <summary>
        /// Static constructor.
        /// </summary>
        static PropertyViewModelFactory()
        {
        }

        /// <summary>
        /// Default constructor.
        /// </summary>
        PropertyViewModelFactory()
        {
            mViewModelByDataType = new Dictionary<Type, Type>();
            mDataTemplateByViewModel = new Dictionary<Type, String>();
        }

        /// <summary>
        /// Gets the unique instance of the factory.
        /// </summary>
        public static PropertyViewModelFactory Instance
        {
            get
            {
                return msInstance;
            }
        }

        /// <summary>
        /// Create a view model according to an instance.
        /// </summary>
        /// <param name="pInstance">The instance to introspect.</param>
        /// <returns>The created view model.</returns>
        public List<PropertyViewModel> CreateViewModel(Object pInstance)
        {
            if 
                (pInstance == null)
            {
                return null;
            }

            // First, get all the property decriptors.
            PropertyDescriptorCollection lProperties = TypeDescriptor.GetProperties(pInstance.GetType());
            List<PropertyViewModel> lViewModels = new List<PropertyViewModel>();
            foreach
                (PropertyDescriptor lPropertyDescriptor in lProperties)
            {
                if
                    (lPropertyDescriptor != null)
                {
                    PropertyViewModel lSimpleViewModel = CreateViewModel(pInstance, lPropertyDescriptor);
                    if
                        (lSimpleViewModel != null)
                    {
                        lViewModels.Add(lSimpleViewModel);
                    }
                }
            }

            return lViewModels;
        }

        /// <summary>
        /// Create a view model according to an instance and a property descriptor.
        /// </summary>
        /// <param name="pInstance">The instance to introspect.</param>
        /// <param name="pDescriptor">The descriptor to introspect.</param>
        /// <returns>The created view model.</returns>
        public PropertyViewModel CreateViewModel(Object pInstance, PropertyDescriptor pDescriptor)
        {
            // Try to find a specific view model.
            PropertyViewModel lPropertyViewModel = null;
            if
                (mViewModelByDataType.ContainsKey(pDescriptor.PropertyType))
            {
                try
                {
                    Object[] lParams = { pInstance, pDescriptor };
                    lPropertyViewModel = Activator.CreateInstance(mViewModelByDataType[pDescriptor.PropertyType], lParams) as PropertyViewModel;
                }
                catch
                    (Exception lEx)
                {
                    Console.WriteLine("UpdateContent() : " + lEx.ToString());
                }
            }

            // If no specific view model, create the default one.
            if
                (lPropertyViewModel == null)
            {
                lPropertyViewModel = new PropertyViewModel(pInstance, pDescriptor);
            }
            Console.WriteLine("CreateViewModel " + lPropertyViewModel);


            return lPropertyViewModel;
        }
        
        /// <summary>
        /// This methods is used to register a view model type on a data type.
        /// eg : AddViewModelForType(typeof(ColorPropertyViewModel), typeof(System.Media.Color));
        /// </summary>
        /// <param name="pViewModelType"></param>
        /// <param name="pDataType"></param>
        public Boolean AddViewModelForType(Type pViewModelType, Type pDataType)
        {
            if
                (! mViewModelByDataType.ContainsKey(pDataType))
            {
                mViewModelByDataType.Add(pDataType, pViewModelType);
                return true;
            }
            return false;
        }

        /// <summary>
        /// This methods is used to register a view model type on a data template identfied by its key.
        /// eg : AddTemplate(typeof(BooleanPropertyViewModel), "CheckBoxTemplate");
        /// </summary>
        /// <param name="pViewModelType"></param>
        /// <param name="pDataType"></param>
        public Boolean AddTemplate(Type pViewModelType, String pDataTemplateKey)
        {
            if
                (!mDataTemplateByViewModel.ContainsKey(pViewModelType))
            {
                mDataTemplateByViewModel.Add(pViewModelType, pDataTemplateKey);
                return true;
            }
            return false;
        }

        /// <summary>
        /// This methods is used to add a new attribute to manage.
        /// eg : AddAttribute(typeof(ShowAttribute), "IsVisible", AttributeType.Boolean, 100);
        /// </summary>
        /// <param name="pAttributeType"></param>
        /// <param name="pPropertyName"></param>
        /// <param name="pAttributeParameterType"></param>
        /// <param name="pPriority"></param>
        public Boolean AddAttribute(Type pAttributeType, Type pPropertyName, AttributeType pAttributeParameterType, Int32 pPriority)
        {
            return false;
        }

        /// <summary>
        /// This function is used to select a data template according to an item.
        /// </summary>
        /// <param name="pPropertyViewModel">The property view model.</param>
        /// <param name="pContainer">The container view model.</param>
        /// <returns>The selected data tempate</returns>
        public override DataTemplate SelectTemplate(Object pPropertyViewModel, DependencyObject pContainer)
        {
            PropertyViewModel lPropertyViewModel = pPropertyViewModel as PropertyViewModel;
            if
                (lPropertyViewModel == null)
            {
                throw new ArgumentException("item must be of type Property");
            }


            DataTemplate lTemplate = FindDataTemplate(lPropertyViewModel, pContainer);
            Console.WriteLine("SelectTemplate " + lTemplate);
            return lTemplate;
        }

        /// <summary>
        /// Ths method finds a data template according to the property view model and its element.
        /// </summary>
        /// <param name="pPropertyViewModel">pPropertyViewModel</param>
        /// <returns>The data template if exists, otherwise return a default template.</returns>
        private DataTemplate FindDataTemplate(PropertyViewModel pPropertyViewModel, DependencyObject pContainer)
        {
            // Use the default template (TextBox)
            DataTemplate lTemplate = null;

            // Try to find a specific view model.
            if
                (mDataTemplateByViewModel.ContainsKey(pPropertyViewModel.GetType()))
            {
                try
                {
                    lTemplate = TryToFindDataTemplate(pContainer as FrameworkElement, mDataTemplateByViewModel[pPropertyViewModel.GetType()]);
                }
                catch
                    (Exception lEx)
                {
                    Console.WriteLine("FindDataTemplate() : " + lEx.ToString());
                }
            }

            // TODO NBY : To extend.
            if 
                (lTemplate == null)
            {
                lTemplate = TryToFindDataTemplate(pContainer as FrameworkElement, "DefaultTemplate");   
            }

            return lTemplate;
        }

        /// <summary>
        /// This method tries to find a data template
        /// </summary>
        /// <param name="pElement">The framework element used to select the data template.</param>
        /// <param name="pKey">The key of the element</param>
        /// <returns>The data template if exists, null otherwise.</returns>
        private static DataTemplate TryToFindDataTemplate(FrameworkElement pElement, String pKey)
        {
            DataTemplate lResource = pElement.TryFindResource(pKey) as DataTemplate;
            return lResource;
        }

        #endregion // Methods.
    }
}
